/**
 * @file couple_config.c
 * @brief Structs and interfaces to defined coupling configuration
 *
 * Stores reads in coupling configuration and sets up data structures
 * to do the coupling.
 *
 * @copyright Copyright  (C)  2021 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *
 * @version 1.0
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef COUPLE_CONFIG_H
#define COUPLE_CONFIG_H

#include "fields.h"
#include "grid.h"
#include "component.h"
#include "yac_mpi.h"

/** \example test_couple_config_parallel1.c
 * This example show how to set up a couple configuration. It uses three
 * processes.
 */

/** \example test_couple_config_parallel2.c
 * This example show how to set up a couple configuration. It uses nine
 * processes.
 */

/** \example test_couple_config_parallel3.c
 * This example show how to set up a couple configuration. It uses nine
 * processes.
 */

/** \example test_couple_config_parallel4.c
 * This example show how to set up a couple configuration. It uses nine
 * processes.
 */

struct yac_couple_config;
/**
 * Constructor for the couple configuration
 * @param[in] xml_filename    coupling xml configuration
 * @param[in] schema_filename corresponding ml schema file
 * @param[in] comm            MPI communicator that contains the processes of
 *                            all components that will be registered with this
 *                            coupling configuration
 * @return couple configuration
 */
struct yac_couple_config * yac_couple_config_new(
  const char * xml_filename, const char * schema_filename, MPI_Comm comm);

/**
 * Destructor for the couple configuration
 * @param[in] config couple configuration to be deleted
 */
void yac_couple_config_delete(struct yac_couple_config * config);

/**
 * Definition of job start and end datetime
 * @param[in] config         couple configuration
 * @param[in] start_datetime calendar job start datetime
 * @param[in] end_datetime   calendar job end datetime
 */
void yac_couple_config_def_datetime(
  struct yac_couple_config * config, const char * start_datetime,
  const char * end_datetime );

/**
 * query routine for the start datetime of the job
 * @param[in] config couple configuration
 * @return start datetime
 */
char * yac_couple_config_get_start_datetime(struct yac_couple_config * config);

/**
 * query routine for the end datetime of the job
 * @param[in] config couple configuration
 * @return end datetime
 */
char * yac_couple_config_get_end_datetime(struct yac_couple_config * config);

/**
 * Get the xml configuration data from the couple configuration
 * @param[in] config couple configuration
 * @return xml configuration data
 */
struct config_xml * yac_couple_config_get_config_xml(
  struct yac_couple_config * config);

/**
 * Defines the components for the couple configuration
 * @param[in] config     couple configuration
 * @param[in] comp_names names of components
 * @param[in] num_comps  number of entries in comp_names
 */
void yac_couple_config_def_components(
  struct yac_couple_config * config,
  char const ** comp_names, size_t num_comps);

/**
 * Adds a coupling field to the couple configuration
 * @param[in] config            couple configuration
 * @param[in] field_name        name of the coupling field
 * @param[in] component_index   component index
 * @param[in] grid              grid
 * @param[in] interp_fields     interpolation fields
 * @param[in] num_interp_fields number of entries in interp_fields
 * @return pointer to coupling field
 */
struct coupling_field * yac_couple_config_add_field(
  struct yac_couple_config * config, char const * field_name,
  size_t component_index, struct yac_basic_grid * grid,
  struct interp_field * interp_fields, size_t num_interp_fields);

/**
 * Gets a component defined in the coupling configuration
 * @param[in] config            couple configuration
 * @param[in] component_index   component index
 * @return component
 */
struct component * yac_couple_config_get_component(
  struct yac_couple_config * config, size_t component_index);

/**
 * Get the coupling fields defined in the couple configuration
 * @param[in]  config     couple configuration
 * @param[out] fields     array containing pointers to coupling fields
 * @param[out] num_fields number of entries in fields
 */
void yac_couple_config_get_fields(
  struct yac_couple_config * config,
  struct coupling_field *** fields, size_t * num_fields);

/**
 * reads coupling configuration from configuration xml file and initiates
 * the generation of all data structures required for interpolation
 * @param[in] config couple   configuration
 */
void yac_setup_couple_config(struct yac_couple_config * config);

/**
 * returns a communicator for communicating between pairs of components
 *
 * @param[in]  config     couple configuration
 * @param[in]  comp_name1 name of first component
 * @param[in]  comp_name2 name of second component
 */
MPI_Comm yac_couple_config_get_pair_rootcomm(
  struct yac_couple_config * config,
  char const * comp_name1, char const * comp_name2 );

#endif // COUPLE_CONFIG_H

