/**
 * @file test_couple_config_parallel4.c
 *
 * @copyright Copyright  (C)  2020 DKRZ, MPI-M
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include <mpi.h>

#include "tests.h"
#include "test_common.h"
#include "xml/config_xml.h"
#include "couple_config.h"
#include "utils.h"
#include "yac_interface.h"
#include "yac_mpi.h"
#include "yaxt.h"
#include "yac_mpi.h"

static char const * component_names[3] = {"comp_1", "comp_2", "comp_3"};

void check_comm(
  struct yac_couple_config * config,
  int comp_1_idx, int comp_2_idx, int ref_size);

int main (void) {

  yac_mpi_call(MPI_Init(NULL, NULL), MPI_COMM_WORLD);

  xt_initialize(MPI_COMM_WORLD);

  int rank, size;
  yac_mpi_call(MPI_Comm_rank(MPI_COMM_WORLD, &rank), MPI_COMM_WORLD);
  yac_mpi_call(MPI_Comm_size(MPI_COMM_WORLD, &size), MPI_COMM_WORLD);

  if (size != 9) {

    PUT_ERR("ERROR: wrong number of processes\n");
    return TEST_EXIT_CODE;
  }

  { // three component, defined on disjoint sets of processes
    struct yac_couple_config * config =
      yac_couple_config_new(
        "couple_config_test_4.xml", "coupling.xsd", MPI_COMM_WORLD);

    yac_couple_config_def_components(
      config, &(component_names[rank / 3]), 1);

    // root comm between comp_1 and comp_2
    if (rank < 6) check_comm(config, 0, 1, 6);

    // root comm between comp_1 and comp_3
    if ((rank < 3) || (rank >= 6)) check_comm(config, 0, 2, 6);

    // root comm between comp_2 and comp_3
    if (rank >= 3) check_comm(config, 1, 2, 6);

    yac_couple_config_delete(config);
  }

  { // two components sharing all processes
    struct yac_couple_config * config =
      yac_couple_config_new(
        "couple_config_test_4.xml", "coupling.xsd", MPI_COMM_WORLD);

    char const * comp_names[2] = {component_names[0], component_names[1]};
    yac_couple_config_def_components(config, comp_names, 2);

    check_comm(config, 0, 1, 9);

    yac_couple_config_delete(config);
  }

  { // three components one is defined on all processes, the other two are
    // a subset
    struct yac_couple_config * config =
      yac_couple_config_new(
        "couple_config_test_4.xml", "coupling.xsd", MPI_COMM_WORLD);

    int num_comps;
    char const * comp_names[2];

    comp_names[0] = component_names[0];
    if (rank < 3) {
      comp_names[1] = component_names[1];
      num_comps = 2;
    } else if (rank < 6) {
      num_comps = 1;
    } else {
      comp_names[1] = component_names[2];
      num_comps = 2;
    }

    yac_couple_config_def_components(config, comp_names, num_comps);

    // root comm between comp_1 and comp_2
    check_comm(config, 0, 1, 9);

    // root comm between comp_1 and comp_3
    check_comm(config, 0, 2, 9);

    // root comm between comp_2 and comp_3
    if ((rank < 3) || (rank >= 6)) check_comm(config, 1, 2, 6);

    yac_couple_config_delete(config);
  }

  { // three components, one having its one processes, the other two sharing
    // some
    struct yac_couple_config * config =
      yac_couple_config_new(
        "couple_config_test_4.xml", "coupling.xsd", MPI_COMM_WORLD);

    int num_comps = 0;
    char const * comp_names[2];

    if (rank < 3)      comp_names[num_comps++] = component_names[0];
    else if (rank < 8) comp_names[num_comps++] = component_names[1];
    if (rank >= 6) comp_names[num_comps++] = component_names[2];

    yac_couple_config_def_components(config, comp_names, num_comps);

    // root comm between comp_1 and comp_2
    if (rank < 8) check_comm(config, 0, 1, 8);

    // root comm between comp_1 and comp_3
    if ((rank < 3) || (rank >= 6)) check_comm(config, 0, 2, 6);

    // root comm between comp_2 and comp_3
    if (rank >= 3) check_comm(config, 1, 2, 6);

    yac_couple_config_delete(config);
  }

  xt_finalize();
  MPI_Finalize();

  return TEST_EXIT_CODE;
}

void check_comm(
  struct yac_couple_config * config,
  int comp_1_idx, int comp_2_idx, int ref_size) {

  MPI_Comm rootcomm =
    yac_couple_config_get_pair_rootcomm(
      config, component_names[comp_1_idx], component_names[comp_2_idx]);
  int rootcomm_size;
  yac_mpi_call(MPI_Comm_size(rootcomm, &rootcomm_size), MPI_COMM_WORLD);

  if (rootcomm_size != ref_size) PUT_ERR("invalid size of rootcomm");

  yac_mpi_call(MPI_Comm_free(&rootcomm), MPI_COMM_WORLD);
}
