/**
 * @file test_point_in_cell.c
 *
 * @copyright Copyright  (C)  2013 DKRZ, MPI-M
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler  <rene.redler@mpimet.mpg.de>
 *
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "tests.h"
#include "grid.h"
#include "geometry.h"
#include "test_common.h"

int main (void) {

   enum yac_edge_type gc_edges[8] =
     {GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE,
      GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE};
   enum yac_edge_type latlon_edges[4] =
     {LAT_CIRCLE, LON_CIRCLE, LAT_CIRCLE, LON_CIRCLE};

   double point_3d[3];
   struct grid_cell cell;

   { // simple inside test
      LLtoXYZ(0, 0, point_3d);
      cell = generate_cell_deg((double[]){-1,1,1,-1},
                               (double[]){-1,-1,1,1}, gc_edges, 4);

      if (!yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (simple inside test)\n")
      yac_free_grid_cell(&cell);
   }

   { // simple outside test
      LLtoXYZ_deg(0, 1.5, point_3d);
      cell = generate_cell_deg((double[]){-1,1,1,-1},
                               (double[]){-1,-1,1,1}, gc_edges, 4);

      if (yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (simple outside test)\n")
      yac_free_grid_cell(&cell);
   }

   { // test point on edge
      LLtoXYZ_deg(0, 1, point_3d);
      cell = generate_cell_deg((double[]){-1,1,1,-1},
                               (double[]){-1,-1,1,1}, gc_edges, 4);
      if (!yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (test point on edge)\n")
      yac_free_grid_cell(&cell);
   }

   { // test near the pole
      LLtoXYZ_deg(22.5, 85.1, point_3d);
      cell = generate_cell_deg((double[]){0,45,45,0},
                               (double[]){85,85,88,88}, latlon_edges, 4);

      if (!yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (lon-lat cell near the pole)\n")
      yac_free_grid_cell(&cell);
   }

   { // test near the pole
      LLtoXYZ_deg(22.5, 85.1, point_3d);
      cell = generate_cell_deg((double[]){0,45,45,0},
                               (double[]){85,85,88,88}, gc_edges, 4);

      if (yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (great circle cell near the pole)\n")
      yac_free_grid_cell(&cell);
   }

   { // simple test at the pole
      LLtoXYZ_deg(45, 88.3, point_3d);
      cell = generate_cell_deg((double[]){0,90,180,270},
                               (double[]){88,88,88,88},
                               (enum yac_edge_type[]){
                                  LAT_CIRCLE, LAT_CIRCLE,
                                  LAT_CIRCLE, LAT_CIRCLE}, 4);

      if (!yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (lon-lat cell on the pole)\n")
      yac_free_grid_cell(&cell);
   }

   { // simple test at the pole
      LLtoXYZ_deg(45, 88.3, point_3d);
      cell = generate_cell_deg((double[]){0,90,180,270},
                               (double[]){88,88,88,88}, gc_edges, 4);

      if (yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (great circle cell on the pole)\n")
      yac_free_grid_cell(&cell);
   }

   { // test point on cell corner and edge
      LLtoXYZ_deg(1, 1, point_3d);
      cell = generate_cell_deg((double[]){0,1,1,0},
                               (double[]){0,0,1,1}, latlon_edges, 4);

      if (!yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (point on corner of lon-lat cell)\n")

      LLtoXYZ_deg(0.5, 0, point_3d);

      if (!yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (point on edge of lon-lat cell)\n")
      yac_free_grid_cell(&cell);
   }

   { // test point on cell corner
      LLtoXYZ_deg(1, 1, point_3d);
      cell = generate_cell_deg((double[]){0,1,0,-1},
                               (double[]){0,1,2,1}, gc_edges, 4);

      if (!yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (point on corner of great circle cell)\n")

      LLtoXYZ_deg(0, 2, point_3d);

      if (!yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (point on corner of great circle cell)\n")

      LLtoXYZ_deg(0, 1, point_3d);

      if (!yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (point in cell)\n")

      LLtoXYZ_deg(0, 3, point_3d);

      if (yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell (point in cell)\n")
      yac_free_grid_cell(&cell);
   }

   { // test point outside of cell
      LLtoXYZ(1.5, 0.5, point_3d);
      cell = generate_cell_deg((double[]){0,1,1,0},
                               (double[]){0,0,1,1}, latlon_edges, 4);

      if (yac_point_in_cell(point_3d, cell))
         PUT_ERR("error in point_in_cell\n")
      yac_free_grid_cell(&cell);
   }

   { // test point on longitude edge of cell


      for (unsigned i = 0; i < 2; ++i) {
         for (int j = 0; j < 3; ++j) {

            enum yac_edge_type edge_type[2][3] =
               {{GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE},
                {LON_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE}};

            double point_lat[3] = {40, 45, 50};

            LLtoXYZ_deg(0.0, point_lat[j], point_3d);
            cell = generate_cell_deg((double[]){0,0,5},
                                     (double[]){40,50,45}, edge_type[i], 3);

            if (!yac_point_in_cell(point_3d, cell))
               PUT_ERR("error in point_in_cell\n")
            yac_free_grid_cell(&cell);
         }
      }
   }

   { // test point on latitude edge of cell

      for (unsigned i = 0; i < 2; ++i) {

         LLtoXYZ_deg(5.0, 10, point_3d);
         double y_coords[2][3] = {{10,10,15}, {10,10,5}};
         cell = generate_cell_deg((double[]){0,10,0}, y_coords[i],
                                  (enum yac_edge_type[])
                                    {LAT_CIRCLE, GREAT_CIRCLE,
                                     GREAT_CIRCLE}, 3);

         if (!yac_point_in_cell(point_3d, cell))
            PUT_ERR("error in point_in_cell\n")
         yac_free_grid_cell(&cell);
      }
   }

   { // test point equator edge of cell

      for (unsigned i = 0; i < 2; ++i) {

         LLtoXYZ_deg(5.0, 0, point_3d);
         double y_coords[2][3] = {{0,0,5}, {0,0,-5}};
         cell = generate_cell_deg((double[]){0,10,0},
                                  y_coords[i], gc_edges, 3);

         if (!yac_point_in_cell(point_3d, cell))
            PUT_ERR("error in point_in_cell\n")
         yac_free_grid_cell(&cell);
      }
   }

   { // test point outside of cell


      for (unsigned i = 0; i < 2; ++i) {

         LLtoXYZ_deg(0.0, 11, point_3d);
         enum yac_edge_type edge_type[2][3] =
            {{LON_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE},
             {GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE}};
         cell = generate_cell_deg((double[]){0,0,5},
                                  (double[]){5,10,20},
                                  edge_type[i], 3);

         if (yac_point_in_cell(point_3d, cell))
            PUT_ERR("error in point_in_cell\n")
         yac_free_grid_cell(&cell);
      }
   }

   { // test point inside of concave cell

      for (unsigned i = 0; i < 2; ++i) {

         LLtoXYZ_deg(0.0, 7.0, point_3d);
         enum yac_edge_type edge_type[2][4] =
            {{LON_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE},
             {GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE}};
         cell = generate_cell_deg((double[]){0, 0, -5, 5},
                                  (double[]){0, 5, 10, 10},
                                  edge_type[i], 4);

         if (!yac_point_in_cell(point_3d, cell))
            PUT_ERR("error in point_in_cell\n")
         yac_free_grid_cell(&cell);
      }
   }

   { // test point outside of concave cell


      for (unsigned i = 0; i < 2; ++i) {

         LLtoXYZ_deg(-2.5, 6.0, point_3d);
         enum yac_edge_type edge_type[2][4] =
            {{LON_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE},
             {GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE, GREAT_CIRCLE}};
         cell = generate_cell_deg((double[]){0, 0, -5, 5},
                                  (double[]){0, 5, 10, 10},
                                  edge_type[i], 4);

         if (yac_point_in_cell(point_3d, cell))
            PUT_ERR("error in point_in_cell\n")
         yac_free_grid_cell(&cell);
      }
   }

   { // from Renes clipping tests

      double x_coords[] = { 0.0, 0.5, 0.0,-0.5};
      double y_coords[] = {-0.5, 0.0, 0.5, 0.0};
      cell = generate_cell_deg(x_coords, y_coords, gc_edges, 4);

      for (unsigned i = 0; i < 4; ++i) {
         LLtoXYZ_deg(x_coords[i], y_coords[i], point_3d);
         if (!yac_point_in_cell(point_3d, cell))
            PUT_ERR("error in point_in_cell\n")
      }
      yac_free_grid_cell(&cell);
   }

   { // example from toy

      cell = generate_cell_rad((double[]){-2.4523357834544868,
                                          -2.5132741228718340,
                                          -2.5742124622891822},
                               (double[]){-0.8955349680874743,
                                          -0.8292526951666540,
                                          -0.8955349680874748},
                               gc_edges, 3);

      LLtoXYZ(-2.5132741229357842, -0.8511807007280388, point_3d);
      if (!yac_point_in_cell(point_3d, cell))
            PUT_ERR("error in point_in_cell\n");
      yac_free_grid_cell(&cell);
   }

   { // example from scrip toy

      cell = generate_cell_3d((double[][3]){{-0.98109801397208807,
                                             -1.4373158006630782e-09,
                                             -0.19351146472502487},
                                            {-0.98296571213463857,
                                             -0.016290939681434788,
                                             -0.18306560040580719},
                                            {-0.98550150905720235,
                                             -0.016447047624685553,
                                             -0.16886761166786302},
                                            {-0.98739170499602125,
                                             -1.9584538881393463e-09,
                                             -0.15829599143708675},
                                            {-0.98550150905854683,
                                             0.016447044383749162,
                                             -0.16886761197567171},
                                            {-0.98296571214686901,
                                             0.016290936203869018,
                                             -0.18306560064960375}},
                               gc_edges, 6);

      if (!yac_point_in_cell((double[]){-0.98437791422272558, 
                                        -1.2055152618042087e-16,
                                        -0.17606851504603632}, cell))
            PUT_ERR("error in point_in_cell\n");
      yac_free_grid_cell(&cell);
   }

   return TEST_EXIT_CODE;
}

